unit uArbiter;

interface

uses
  uVect, uBody, uShape, math;

type
  TContact = record
    p, n: cpVect;
    dist: single;
    r1, r2: cpVect;
    nMass, tMass, bounce: single;
    jnAcc, jtAcc, jBias: single;
    bias: single;
    hash: integer;
  end;

type
  cpContacts = array of TContact;

type
  PArbiter = ^TArbiter;
  TArbiter = record
	  numContacts: integer;
	  contacts: cpContacts;
  	shape1, shape2: PShape;
	  u, e: single;
	  target_v: cpVect;
    active: boolean;
    passivetime: word;
    Next, Prev: PArbiter;
  end;

var
  cpArbiter: PArbiter;
  ArbitersCount: integer;

  function cpContactInit(p, n: cpVect; dist: single; hash: integer): TContact;
  function cpArbiterCreate(shape1, shape2: PShape): PArbiter;
  procedure cpArbiterDelete(arb: PArbiter);
  procedure cpArbiterInject(arb: PArbiter; contacts: cpContacts; numContacts: integer);
  procedure cpArbiterPreStep(arb: PArbiter; dt_inv: single);
  procedure cpArbiterApplyImpulse(arb: PArbiter);

implementation

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpContactInit(p, n: cpVect; dist: single; hash: integer): TContact;
var
  con: TContact;
begin
	con.p:= p;
	con.n:= n;
	con.dist:= dist;
	
	con.jnAcc:= 0;
	con.jtAcc:= 0;
	con.jBias:= 0;

	con.hash:= hash;
  Result:= con;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpArbiterCreate(shape1, shape2: PShape): PArbiter;
var
  arb: PArbiter;
begin
  New(arb);
  arb.Next:= nil;
  arb.Prev:= nil;

  arb.numContacts:= 0;
  arb.shape1:= shape1;
  arb.shape2:= shape2;
  arb.passivetime:= 0;

  arb.Next := cpArbiter;
  if cpArbiter <> nil then
    cpArbiter.Prev := arb;
  cpArbiter := arb;

  Result := cpArbiter;
  inc(ArbitersCount);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpArbiterDelete(arb: PArbiter);
var
  DelArb: PArbiter;
begin
  DelArb:= arb;

  SetLength(arb.contacts, 0);
  arb.shape1:= nil;
  arb.shape2:= nil;

  if arb.Prev <> nil then arb.Prev.Next:= arb.Next
    else cpArbiter:= arb.Next;
  if arb.Next <> nil then arb.Next.Prev:= arb.Prev;

  Dispose(DelArb);
  dec(ArbitersCount);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpArbiterInject(arb: PArbiter; contacts: cpContacts; numContacts: integer);
var
  i, j: integer;
begin
  for i:= 0 to arb.numContacts - 1 do
    for j:= 0 to numContacts - 1 do
      if contacts[j].hash = arb.contacts[i].hash then
      begin
        contacts[j].jnAcc:= arb.contacts[i].jnAcc;
        contacts[j].jtAcc:= arb.contacts[i].jtAcc;
      end;

  arb.active:= true;
  arb.numContacts:= numContacts;
  SetLength(arb.contacts, numContacts);
  for i:= 0 to numContacts - 1 do
    arb.contacts[i]:= contacts[i];
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpArbiterPreStep(arb: PArbiter; dt_inv: single);
const
  cp_bias_coef = 0.1;
  cp_collision_slop = 0.1;
var
  s1, s2: PShape;
  b1, b2: PBody;
  i: integer;
  con: TContact;
  mass_sum, r1cn, r2cn, kn, r1ct, r2ct, kt: single;
  t, v1, v2, j: cpVect;
begin
  s1:= arb.shape1;
  s2:= arb.shape2;

  arb.e:= s1.e * s2.e;
  arb.u:= s1.u * s2.u;
  arb.target_v:= cpvsub(s2.surface_v, s1.surface_v);

  b1:= s1.body;
  b2:= s2.body;

  for i:= 0 to arb.numContacts - 1 do
  begin
    con:= arb.contacts[i];

    con.r1:= cpvsub(con.p, b1.p);
    con.r2:= cpvsub(con.p, b2.p);

		mass_sum:= b1.m_inv + b2.m_inv;

		r1cn:= cpvcross(con.r1, con.n);
		r2cn:= cpvcross(con.r2, con.n);
		kn:= mass_sum + b1.i_inv * r1cn * r1cn + b2.i_inv * r2cn * r2cn;
		con.nMass:= 1 / kn;

		t:= cpvperp(con.n);
		r1ct:= cpvcross(con.r1, t);
		r2ct:= cpvcross(con.r2, t);
		kt:= mass_sum + b1.i_inv * r1ct * r1ct + b2.i_inv * r2ct * r2ct;
		con.tMass:= 1 / kt;
				
		con.bias:= - cp_bias_coef * dt_inv * min(0, con.dist + cp_collision_slop);
		con.jBias:= 0;

		v1:= cpvadd(b1.v, cpvmult(cpvperp(con.r1), b1.w));
		v2:= cpvadd(b2.v, cpvmult(cpvperp(con.r2), b2.w));
		con.bounce:= cpvdot(con.n, cpvsub(v2, v1)) * arb.e;

		j:= cpvadd(cpvmult(con.n, con.jnAcc), cpvmult(t, con.jtAcc));

    arb.contacts[i]:= con;
		cpBodyApplyImpulse(b1, cpvneg(j), con.r1);
		cpBodyApplyImpulse(b2, j, con.r2);
  end;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpArbiterApplyImpulse(arb: PArbiter);
var
  b1, b2: PBody;
  con: TContact;
  i: integer;
  n, r1, r2, vb1, vb2, jb, v1, v2, vr, t, j: cpVect;
  vbn, jbn, jbnOld, vrn, jn, jnOld, vrt, jtMax, jt, jtOld: single;
begin

  b1:= arb.Shape1.Body;
  b2:= arb.Shape2.Body;

  for i:= 0 to arb.numContacts - 1 do
  begin
    con:= arb.contacts[i];
		n:= con.n;
		r1:= con.r1;
		r2:= con.r2;

		vb1:= cpvadd(b1.v_bias, cpvmult(cpvperp(r1), b1.w_bias));
		vb2:= cpvadd(b2.v_bias, cpvmult(cpvperp(r2), b2.w_bias));
		vbn:= cpvdot(cpvsub(vb2, vb1), n);

    jbn:= (con.bias - vbn) * con.nMass;
		jbnOld:= con.jBias;
		con.jBias:= max(jbnOld + jbn, 0);
		jbn:= con.jBias - jbnOld;

		jb:= cpvmult(n, jbn);
		cpBodyApplyBiasImpulse(b1, cpvneg(jb), r1);
		cpBodyApplyBiasImpulse(b2, jb, r2);

		v1:= cpvadd(b1.v, cpvmult(cpvperp(r1), b1.w));
		v2:= cpvadd(b2.v, cpvmult(cpvperp(r2), b2.w));
		vr:= cpvsub(v2, v1);
		vrn:= cpvdot(vr, n);

		jn:= -(con.bounce + vrn) * con.nMass;
		jnOld:= con.jnAcc;
		con.jnAcc:= max(jnOld + jn, 0);
		jn:= con.jnAcc - jnOld;

		t:= cpvperp(n);
		vrt:= cpvdot(cpvadd(vr, arb.target_v), t);

		jtMax:= arb.u * con.jnAcc;
		jt:= -vrt * con.tMass;
		jtOld:= con.jtAcc;
		con.jtAcc:= min(max(jtOld + jt, -jtMax), jtMax);
		jt:= con.jtAcc - jtOld;

    arb.contacts[i]:= con;

		j:= cpvadd(cpvmult(n, jn), cpvmult(t, jt));

		cpBodyApplyImpulse(b1, cpvneg(j), r1);
		cpBodyApplyImpulse(b2, j, r2);
  end;
end;

end.
