unit uBody;

interface

uses
  uVect, math;

type
  PBody = ^TBody;
  TBody = record
	  // Mass and it's inverse.
    m, m_inv: single;
   	// Moment of inertia and it's inverse.
    i, i_inv: single;
    // NOTE: v_bias and w_bias are used internally for penetration/joint correction.
   	// Linear components of motion (position, velocity, and force), Unit length
    p, v, f, v_bias, rot: cpVect;
    // Angular components of motion (angle, angular velocity, and torque)
    a, w, t, w_bias: single;
    Next, Prev: PBody;
  end;

var
  cpBody: PBody;
  BodiesCount: integer;

  function cpBodyCreate(m, i: single): PBody;
  procedure cpBodyDelete(body: PBody);
  procedure cpBodyResetParams(body: PBody);
  procedure cpBodyApplyImpulse(body: PBody; j, r: cpVect);
  procedure cpBodyApplyBiasImpulse(body: PBody; j, r: cpVect);
  procedure cpBodyResetForce(body: PBody);
  procedure cpBodyApplyForce(body: PBody; f: cpVect; r: cpVect);
  procedure cpBodySetMass(body: PBody; m: single);
  procedure cpBodySetMoment(body: PBody; i: single);
  procedure cpBodySetAngle(body: PBody; a: single);
  procedure cpBodyUpdateVelocity(body: PBody; gravity: cpVect; damping, dt: single);
  procedure cpBodyUpdatePosition(body: PBody; dt: single);
  function cpMomentForCircle(m, r1, r2: single; offset: cpVect): single;
  function cpMomentForPoly(m: single; numVerts: integer; const verts: array of cpVect; offset: cpVect): single;
  
implementation

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpBodyCreate(m, i: single): PBody;
var
  b: PBody;
begin
  New(b);
  b.Next:= nil;
  b.Prev:= nil;
  b.p:= cpv(0, 0);

  cpBodyResetparams(b);
  cpBodySetMass(b, m);
	cpBodySetMoment(b, i);
  cpBodySetAngle(b, 0);

  b.Next:= cpBody;
  if cpBody <> nil then cpBody.Prev:= b;
  cpBody:= b;
  Result:= cpBody;

  inc(BodiesCount);
end;

// added by Hex[man] ...
procedure cpBodyResetParams(body: PBody);
begin
  body.v:= cpv(0, 0);
  body.f:= cpv(0, 0);
	body.v_bias:= cpv(0,0);
	body.w:= 0;
	body.t:= 0;
	body.w_bias:= 0;
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpBodyDelete(body: PBody);
var
  DelBody: PBody;
begin
  DelBody:= body;
  if body.Prev <> nil then body.Prev.Next := body.Next
    else cpBody:= body.Next;
  if body.Next <> nil then body.Next.Prev := body.Prev;
  Dispose(DelBody);
  dec(BodiesCount);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpBodyApplyImpulse(body: PBody; j, r: cpVect);
begin
  body.v:= cpvadd(body.v, cpvmult(j, body.m_inv));
  body.w:= body.w + (body.i_inv * cpvcross(r, j));
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpBodyApplyBiasImpulse(body: PBody; j, r: cpVect);
begin
  body.v_bias:= cpvadd(body.v_bias, cpvmult(j, body.m_inv));
  body.w_bias:= body.w_bias + body.i_inv * cpvcross(r, j);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpBodySetMass(body: PBody; m: single);
begin
	body.m:= m;
	body.m_inv:= 1 / m;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpBodySetMoment(body: PBody; i: single);
begin
	body.i:= i;
	body.i_inv:= 1 / i;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpBodySetAngle(body: PBody; a: single);
begin
  body.a:= (pi * 2) * frac(a / (pi * 2));
  body.rot:= cpvforangle(a);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpBodyUpdateVelocity(body: PBody; gravity: cpVect; damping, dt: single);
begin
  body.v:= cpvadd(cpvmult(body.v, damping), cpvmult(cpvadd(gravity, cpvmult(body.f, body.m_inv)), dt));
  body.w:= body.w * damping + body.t * body.i_inv * dt;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpBodyUpdatePosition(body: PBody; dt: single);
begin
  body.p:= cpvadd(body.p, cpvmult(cpvadd(body.v, body.v_bias), dt));
  cpBodySetAngle(body, body.a + (body.w + body.w_bias) * dt);
  body.v_bias:= cpv(0, 0);
  body.w_bias:= 0;
end;
//--------------------------------
//--------------------------------
procedure cpBodyResetForce(body: PBody);
begin
  body.f:=cpv(0,0);
  body.t:=0.0;
end;
//--------------------------------
//--------------------------------
procedure cpBodyApplyForce(body: PBody; f: cpVect; r: cpVect);
begin
  body.f:=cpvadd(body.f,f);
  body.t:=body.t+cpvcross(r,f);
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpMomentForCircle(m, r1, r2: single; offset: cpVect): single;
begin
	Result:= (1 / 2) * m * (r1 * r1 + r2 * r2) + m * cpvdot(offset, offset);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpMomentForPoly(m: single; numVerts: integer; const verts: array of cpVect; offset: cpVect): single;
var
  i: integer;
  tVerts: array of cpVect;
  sum1, sum2, a, b: single;
  v1, v2: cpVect;
begin
  SetLength(tVerts, numVerts);
	for i:= 0 to numVerts - 1 do
		tVerts[i]:= cpvadd(verts[i], offset);
	
	sum1:= 0;
	sum2:= 0;
	for i:= 0 to numVerts - 1 do
  begin
		v1:= tVerts[i];
		v2:= tVerts[(i+1) mod numVerts];

		a:= cpvcross(v2, v1);
		b:= cpvdot(v1, v1) + cpvdot(v1, v2) + cpvdot(v2, v2);
		
		sum1:= sum1 + a * b;
		sum2:= sum2 + a;
	end;
	
	Result:= (m * sum1) / (6 * sum2);
end;

end.
