unit uShape;

interface

uses
  uVect, uBody, uBB;

type
  cpShapeType = (CP_CIRCLE_SHAPE, CP_SEGMENT_SHAPE, CP_POLY_SHAPE);

type
  PCircleShape = ^TCircleShape;
  TCircleShape = record
	  c: cpVect;
	  r: single;
	  tc: cpVect;
  end;

type
  PSegmentShape = ^TSegmentShape;
  TSegmentShape = record
    a, b, n: cpVect;
    r: single;
    ta, tb, tn: cpVect;
  end;

type
  PShape = ^TShape;
  TShape = record
    ShapeType: cpShapeType;
    bb: cpBB;
    group, layers, id: integer;
    //elasticity, friction
    e, u: single;
    surface_v: cpVect;
    Body: PBody;
    DfShape: pointer;
    Next, Prev: PShape;
  end;
var
  cpShape: PShape;
  ShapesCount: integer;
  SHAPE_ID_COUNTER: integer;

  procedure updateBB(s: PShape);
  function cpShapeCreate(ShapeType: cpShapeType; body: PBody): PShape;
  procedure cpShapeDelete(s: PShape);
  function cpCircleShapeCreate(body: PBody; radius: single; offset: cpVect): PShape;
  function cpSegmentShapeCreate(body: PBody; a, b: cpVect; r: single): PShape;

implementation

uses
  uPolyShape;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpCircleShapeCacheData(circ: PCircleShape; p, rot: cpVect): cpBB;
var
  c: cpVect;
  r: single;
begin
	circ.tc:= cpvadd(p, cpvrotate(circ.c, rot));
  c:= circ.tc;
  r:= circ.r;
	Result:= cpBBNew(c.x - r, c.y - r, c.x + r, c.y + r);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpSegmentShapeCacheData(seg: PSegmentShape; p, rot: cpVect): cpBB;
var
  l, r, s, t, rad: single;
begin
	seg.ta:= cpvadd(p, cpvrotate(seg.a, rot));
	seg.tb:= cpvadd(p, cpvrotate(seg.b, rot));
	seg.tn:= cpvrotate(seg.n, rot);

	if seg.ta.x < seg.tb.x then
  begin
		l:= seg.ta.x;
		r:= seg.tb.x;
	end else
  begin
		l:= seg.tb.x;
		r:= seg.ta.x;
	end;
	
	if seg.ta.y < seg.tb.y then
  begin
		s:= seg.ta.y;
		t:= seg.tb.y;
	end else
  begin
		s:= seg.tb.y;
		t:= seg.ta.y;
	end;

	rad:= seg.r;
	Result:= cpBBNew(l - rad, s - rad, r + rad, t + rad);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure updateBB(s: PShape);
begin
  if s.ShapeType = CP_CIRCLE_SHAPE then s.bb:= cpCircleShapeCacheData(s.DfShape, s.Body.p, s.Body.rot);
  if s.ShapeType = CP_SEGMENT_SHAPE then s.bb:= cpSegmentShapeCacheData(s.DfShape, s.Body.p, s.Body.rot);
  if s.ShapeType = CP_POLY_SHAPE then s.bb:= cpPolyShapeCacheData(s.DfShape, s.Body.p, s.Body.rot);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpShapeCreate(ShapeType: cpShapeType; body: PBody): PShape;
var
  s: PShape;
begin
  New(s);
  s.Next:= nil;
  s.Prev:= nil;

  s.ShapeType:= ShapeType;
  s.Body:= body;
  s.layers:= 0;
  s.group:= 0;
  s.e:= 0;
  s.u:= 0;
  s.surface_v:= cpv(0, 0);
  s.id:= SHAPE_ID_COUNTER;    
  inc(SHAPE_ID_COUNTER);

  s.Next:= cpShape;
  if cpShape <> nil then cpShape.Prev:= s;
  cpShape:= s;
  Result:= cpShape;

  inc(ShapesCount);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpShapeDelete(s: PShape);
var
  DelShape: PShape;
begin
  DelShape:= s;
  if s.Prev <> nil then s.Prev.Next := s.Next
    else cpShape:= s.Next;
  if s.Next <> nil then s.Next.Prev := s.Prev;

  s.Body:= nil;
  Dispose(s.DfShape);
  Dispose(DelShape);
  dec(BodiesCount);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpCircleShapeCreate(body: PBody; radius: single; offset: cpVect): PShape;
var
  s: PShape;
  sd: PCircleShape;
begin
  New(sd);


  s:= cpShapeCreate(CP_CIRCLE_SHAPE, body);
  s.DfShape:= sd;
  sd.r:= radius;
  sd.c:= offset;

  updateBB(s);

  Result:= s;
end;


//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
function cpSegmentShapeCreate(body: PBody; a, b: cpVect; r: single): PShape;
var
  s: PShape;
  sd: PSegmentShape;
begin
  New(sd);

  s:= cpShapeCreate(CP_SEGMENT_SHAPE, body);
  s.DfShape:= sd;
  s.body:= body;
  
  sd.a:= a;
	sd.b:= b;
  sd.n:= cpvperp(cpvnormalize(cpvsub(b, a)));
  sd.r:= r;

  updateBB(s);

  Result:= s;
end;

end.
