unit uSpace;

interface

uses
  uBody, uShape, uArbiter, uCollision, uJoint, uVect, uBB, math, windows,sNail;

type
  cpSpace = record
    iterations: integer;
    gravity: cpVect;
    damping: single;
  end;

var
  space: cpSpace;

  procedure cpSpaceInit;
  procedure cpSpaceStep(dt: single);
  procedure cpSpaceClear;
  procedure DeleteBody(b: PBody);
  
implementation

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpSpaceInit;
begin
	space.iterations:= 10;
  space.damping:= 1;
  log_AddStr(nil, '- Chipmunk > Initialized Successfully');
end;


//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

function queryReject(s1, s2: PShape): boolean;

begin

  Result:= false;
  if s1.layers <> s2.layers then exit;
  if s1.body = s2.body then exit;
  if (s1.group = s2.group) and (s1.group <> 0) and (s2.group <> 0) then exit;
  if not(cpBBintersects(s1.bb, s2.bb)) then exit;
  Result:= true;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure queryFunc(s1, s2: PShape);
var
  arb, arbN: PArbiter;
  numContacts: integer;
  contacts: cpContacts;
  s: PShape;
begin

  if not queryReject(s1, s2) then exit;

  if s1.ShapeType > s2.ShapeType then
  begin
    s:= s1;
    s1:= s2;
    s2:= s;
  end;

  numContacts:= cpCollideShapes(s1, s2, contacts);
  if numContacts = 0 then exit;

  arbN:= nil;
  arb:= cpArbiter;
  while arb <> nil do
  begin
		if ((arb.shape1 = s1) and (arb.shape2 = s2)) or ((arb.shape1 = s2) and (arb.shape2 = s1)) then
    begin
      arbN:= arb;
      break;
    end;
    arb:= arb.Next;
  end;

  if arbN = nil then arbN:= cpArbiterCreate(s1, s2);
  cpArbiterInject(arbN, contacts, numContacts);
end;


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpSpaceStep(dt: single);
var
  i, j, k: integer;
  dt_inv, damping: single;
  arb, arbNext: PArbiter;
  b: PBody;
  s: PShape;
  jnt: PJoint;
  StaticShapes, DynamicShapes: array of PShape;
  StaticShapesCount, DynamicShapesCount: integer;
begin
	if dt = 0 then exit;
	dt_inv:= 1 / dt;

  arb:= cpArbiter;
  while arb <> nil do
  begin
    arbNext:= arb.Next;
		arb.active:= false;
    if arb.passivetime > 5000 then cpArbiterDelete(arb);
    arb:= arbNext;
  end;

	damping:= Power(1 / space.damping, -dt);

  b:= cpBody;
  while b <> nil do
  begin
    if b.m <> INFINITY then cpBodyUpdateVelocity(b, space.gravity, damping, dt);

    b:= b.next;
  end;

  StaticShapesCount:= 0;
  DynamicShapesCount:= 0;
  s:= cpShape;
  while s <> nil do
  begin
	  updateBB(s);

    if s.Body.m = INFINITY then
    begin
      StaticShapesCount:= StaticShapesCount + 1;
      SetLength(StaticShapes, StaticShapesCount);
      StaticShapes[StaticShapesCount - 1]:= s;
    end else
    begin
      DynamicShapesCount:= DynamicShapesCount + 1;
      SetLength(DynamicShapes, DynamicShapesCount);
      DynamicShapes[DynamicShapesCount - 1]:= s;
    end;
    s:= s.next;
  end;

  for i:= 0 to DynamicShapesCount - 1 do
    for j:= 0 to StaticShapesCount - 1 do
      queryFunc(DynamicShapes[i], StaticShapes[j]);

  if DynamicShapesCount > 1 then
    for i:= 0 to DynamicShapesCount - 2 do
      for j:= i + 1 to DynamicShapesCount - 1 do
        queryFunc(DynamicShapes[i], DynamicShapes[j]);

  arb:= cpArbiter;
  while arb <> nil do
  begin
		if arb.active = true then cpArbiterPreStep(arb, dt_inv);
    if arb.active = true then arb.passivetime:= 0 else arb.passivetime:= arb.passivetime + 1;
    arb:= arb.Next;
  end;

  jnt:= cpJoint;
  while jnt <> nil do
  begin
	  JointPreStep(jnt, dt_inv);
    jnt:= jnt.Next;
  end;

  for k:= 0 to space.iterations - 1 do
  begin
    arb:= cpArbiter;
    while arb <> nil do
    begin
      if arb.active = true then cpArbiterApplyImpulse(arb);
      arb:= arb.Next;
    end;

    jnt:= cpJoint;
    while jnt <> nil do
    begin
      JointApplyImpulse(jnt);
      jnt:= jnt.Next;
    end;
  end;

  b:= cpBody;
  while b <> nil do
  begin
    cpBodyUpdatePosition(b, dt);
    b:= b.Next;
  end;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpSpaceClear;
var
  arb, arbNext: PArbiter;
  s, sNext: PShape;
  jnt, jntNext: PJoint;
  b, bNext: PBody;
begin
  arb:= cpArbiter;
  while arb <> nil do
  begin
    arbNext:= arb.Next;
    cpArbiterDelete(arb);
    arb:= arbNext;
  end;

  jnt:= cpJoint;
  while jnt <> nil do
  begin
    jntNext:= jnt.Next;
    cpJointDelete(jnt);
    jnt:= jntNext;
  end;

  s:= cpShape;
  while s <> nil do
  begin
    sNext:= s.Next;
    cpShapeDelete(s);
    s:= sNext;
  end;

  b:= cpBody;
  while b <> nil do
  begin
    bNext:= b.Next;
    cpBodyDelete(b);
    b:= bNext;
  end;

  log_AddStr(nil, '- Chipmunk > Resources cleaned');
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure DeleteBody(b: PBody);
var
  arb, arbNext: PArbiter;
  s, sNext: PShape;
  jnt, jntNext: PJoint;
begin
  arb:= cpArbiter;
  while arb <> nil do
  begin
    arbNext:= arb.Next;
    if (arb.shape1.Body = b) or (arb.shape2.Body = b) then cpArbiterDelete(arb);
    arb:= arbNext;
  end;

  jnt:= cpJoint;
  while jnt <> nil do
  begin
    jntNext:= jnt.Next;
    if (jnt.body1 = b) or (jnt.body2 = b) then cpJointDelete(jnt);
    jnt:= jntNext;
  end;

  s:= cpShape;
  while s <> nil do
  begin
    sNext:= s.Next;
    if s.Body = b then cpShapeDelete(s);
    s:= sNext;
  end;

  cpBodyDelete(b);
end;

end.


